;; $Id: table.dsl,v 1.5 1998/12/14 13:33:56 tkg Exp $
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; MODULE:    Table Stylesheet
;; VERSION:   1.0
;; DATE:      May 1, 1998
;;
;; MULBERRY INTERNAL VERSION CONTROL:
;; $Id: table.dsl,v 1.5 1998/12/14 13:33:56 tkg Exp $
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SYSTEM:    Mulberry DSSSL Processing
;;
;; PURPOSE:   This is the Table module of the DSSSL stylesheet for
;;            use in Mulberry Tag Libraries
;;
;; CONTAINS:  1) Definition of parameters
;;            2) TABLE element construction rule
;;            3) Functions used in element construction rules
;;            4) Element construction rules for other elements
;;            5) Additional procedures
;;
;; PACKAGES REQUIRED:
;;            None
;;
;; ORIGINAL CREATION DATE:
;;            January 1997
;;
;; CREATED BY:
;;            Mulberry Technologies, Inc.
;;            17 West Jefferson Street, Suite 207
;;            Rockville, MD  20850
;;            Phone:  301/315-9631
;;            Fax:    301/315-8285
;;            email:  info@mulberrytech.com
;;            WWW:    http://www.mulberrytech.com
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Table code borrowed from Jon Bosak's db63.dsl DocBook style sheet
;; as modified

;; 961028 -- bosak
;;   TABLE element spec changed
;;   p-style changed to para-style
;; 961123 -- bosak
;;   TABLE and TITLE elements modified
;;   margins changed
;;   added INFORMALTABLE (docbook-specific)
;; 961124 -- bosak
;;   mods to THEAD
;; 970116 -- bosak
;;   method for assigning frame-attribute on TGROUP changed from
;;     attribute-string to inherited-attribute string to cope
;;     with FRAME attribute set on either TABLE or INFORMALTABLE
;;     (docbook-specific)
;;   mods to TITLE, cell indents & margins
;; Notes by bosak:
;;   For (informal)tables and table titles to align correctly with
;;     paragraphs immediately preceding them, the (informal)table
;;     must be a child of the preceding paragraph.
;;   To get a table frame and rules around each cell, you must set
;;     frame="all" on every TABLE and INFORMALTABLE and
;;     colsep=1 rowsep=1 on every TGROUP.
;;   All further notes by Anders Berglund except where noted.
;;
;; ============================ TABLES ===============================
;
; *** DRAFT VERSION ****
;
; Copyright (C), Berglund Consulting & Type Foundry 1996.
; Permission to copy in any form is granted for use in  
; DSSSL applications, provided this notice is included in
; all copies.
;
; This supports the specifications in the "Exchange model" in the
; SGML Open Technical Resolution TR 9503:1995
; with the modification that SPANSPECs are supported.
; Thus the following is NOT supported:
; - mixed measure - e.g. 2*+3pt - for colspecs

; Caution
; - NOTE that vertical column spans are not supported by Microsoft Word
; - NOTE that for RTF the table foot is placed at the end of the table;
;   table heads are correctly repeated at page breaks in a table
;
; Tailor these values to go with the rest of the DSSSL application
; and for the desired default values
;
(define %cals-rule-default% 1)
(define %cals-valign-default% "TOP")
; cell margins - a 4pt value assumes that the paragraphs in the
;                cells have a 0 start and end indent so that a  
;                margin needs to be specified on the cells
;; bosak has tweaked all these
(define %cals-cell-before-row-margin% 1.5pt)
(define %cals-cell-after-row-margin% 1.5pt)
(define %cals-cell-before-column-margin% 3pt)
(define %cals-cell-after-column-margin% 3pt)
; value for start and end indent; initial value for inheritance in the
;                                 cells
(define %cals-cell-content-start-indent% 3pt)
(define %cals-cell-content-end-indent% 2pt)
;
;; ----------------------------------------------------------------------
;; Table rule widths

(define %table-before-row-border% 1pt)
(define %table-after-row-border% 1pt)
(define %table-before-column-border% #f) 
(define %table-after-column-border% #f)
(define %table-head-body-border% 1pt)
(define %table-cell-after-column-border% #f)
(define %table-cell-after-row-border% 0.5pt)

;; Color to use for table cell column border
(define %table-cell-border-color% (color rgb-color-space 0.75 0.75 0.75))

;; Functions to hide the sometimes boolean, sometimes length nature of
;; our table rule width variables by defining variables that are
;; either a boolean or a table-border flow object
(define calc-table-before-row-border
  (if (boolean? %table-before-row-border%)
      %table-before-row-border%
      ;; Avoid problems with the DSSSL compiler when 
      ;; %table-before-row-border% is boolean.
      (let ((border-width %table-before-row-border%))
	(make table-border
	  line-thickness: border-width))))

(define calc-table-after-row-border
  (if (boolean? %table-after-row-border%)
      %table-after-row-border%
      (let ((border-width %table-after-row-border%))
	(make table-border
	  line-thickness: border-width))))

(define calc-table-before-column-border
  (if (boolean? %table-before-column-border%)
      %table-before-column-border%
      (let ((border-width %table-before-column-border%))
	(make table-border
	  line-thickness: border-width))))

(define calc-table-after-column-border
  (if (boolean? %table-after-column-border%)
      %table-after-column-border%
      (let ((border-width %table-after-column-border%))
	(make table-border
	  line-thickness: border-width))))

(define calc-table-head-body-border
  (if (boolean? %table-head-body-border%)
      %table-head-body-border%
      (let ((border-width %table-head-body-border%))
	(make table-border
	  line-thickness: border-width))))

(define calc-table-cell-after-column-border
  (if (boolean? %table-cell-after-column-border%)
      %table-cell-after-column-border%
      (let ((border-width %table-cell-after-column-border%))
	(make table-border
	  color: %table-cell-border-color%
	  line-thickness: border-width))))

(define calc-table-cell-after-row-border
  (if (boolean? %table-cell-after-row-border%)
      %table-cell-after-row-border%
      (let ((border-width %table-cell-after-row-border%))
	(make table-border
	  color: %table-cell-border-color%
	  line-thickness: border-width))))

; These may need changing for the desired style
;
(element INFORMALTABLE
  (make display-group
    space-before: %block-sep%
    space-after: %block-sep%
    start-indent: (inherited-start-indent)
    (process-children)))

(define %pgwide-start-indent% 0pt)

(element TABLE
  (make display-group
    space-before: %block-sep%
    space-after: %block-sep%
    start-indent: (let ((pgwide (attribute-string "pgwide")))
		    (if
		     (and
		      (string? pgwide)
		      (not
		       (= (string->number pgwide) 0)))
		     %pgwide-start-indent%
		     (inherited-start-indent)))
    font-size: (let ((pgwide (attribute-string "pgwide")))
		 (if
		  (and
		   (string? pgwide)
		   (not
		    (= (string->number pgwide) 0)))
		  (- %bf-size% 2pt)
		  %bf-size%))
    (with-mode table-caption-mode
      (process-first-descendant "TITLE"))
    (process-children)))

(mode table-caption-mode
  (element TITLE
    (let ((label (attribute-string "label" (ancestor "table"))))
      (make paragraph
	use: figure-title-style
	space-before: (b2b-space 18pt)
	space-after: (b2b-space 18pt)
	start-indent: (+ (inherited-start-indent)
			 %figure-number-indent%)
	first-line-start-indent: (- %figure-number-indent%)
	(make line-field
	  font-weight: 'bold
	  field-width: %figure-number-indent%
	  (literal
	   (string-append
	    "Table "
	    (if label
		label
		(format-number
		 (element-number (parent (current-node)))
		 "1")))))
	(process-children-trim)))))

;---------------------------------------------------------------------
;   
; There should be no need to change the specification below 
;
;---------------------------------------------------------------------

(define (CALS-COLSPEC-UNIT u)
  (if (string? u)
      (let ((strlen (string-length u)))
	(if (string=? "*" (substring u (- strlen 1) strlen)) 
	    (let* ((pnum (substring u 0 (- strlen 1))))
	      (if (number? (string->number pnum))
                  (table-unit (string->number pnum))
                  (table-unit 1)))
	    (if (> strlen 2)
		(let ((u-s-i (UNAME-START-INDEX u (- strlen 1))))
		  (if (= u-s-i 0);; there's no number here
		      1pi;; so return something that might work
		      (if (= u-s-i strlen);; there's no unit name here
			  (* (string->number u) 1pt);; so default to points
			  (let* ((unum (string->number
					(substring u 0 u-s-i)))
				 (uname (STRING-DOWNCASE
					 (substring u u-s-i strlen))))
			    (case uname
                              (("mm") (* unum 1mm))
                              (("cm") (* unum 1cm))
                              (("in") (* unum 1in))
                              (("pi") (* unum 1pi))
                              (("pt") (* unum 1pt))
                              (else
                               (cond 
                                ((number? unum)
                                 (* unum 1pt))
                                ((number? (string->number u))
                                 (* (string->number u) 1pt))
				(else u))))))))
		(if (number? (string->number u))
		    (* (string->number u) 1pt)
		    (table-unit 1)))))
      (table-unit 1)))


; given a node list "nodes" find the snl that has a gi matching "giname"
; and an attribute "attname" that has the value "attval"
; if no such node return #f
(define (GI-ATTVAL-NODE-IN-NODELIST giname attname attval nodes)
  (let* ((n (node-list-first nodes)) ;; has to be let* bosak 961123
	 (attnamestr (attribute-string attname n))) ;; added by bosak 961123
    (if (and (string=? (gi n) giname) attnamestr (string=? attnamestr attval)) ;; check for attnamestr added by bosak 961123
;;    (if (and (string=? (gi n) giname) (string=? (attribute-string attname n) attval)) ;; old version dies if attribute-string returns #f
        n
        (if (node-list-empty? (node-list-rest nodes))
            #f
            (GI-ATTVAL-NODE-IN-NODELIST giname attname attval (node-list-rest nodes))))))

; find the child number of the "colspec" that has a "colname" attribute
; whose value matches the "namest" attribute value of the "spanspec"
; whose "spannane" attribute value matches the "spanname" value of the "entry"
(define (CALS-ENTRY-SPANSPEC-START)
  (child-number
    (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC"  "colname"
      (attribute-string "namest" 
        (GI-ATTVAL-NODE-IN-NODELIST "SPANSPEC" "spanname"
          (attribute-string "spanname") (children (ancestor "tgroup"))))
      (children (ancestor "tgroup")))))

; find the child number of the "colspec" that has a "colname" attribute
; value matching the "namest" attribute value of the "entry"
(define (CALS-ENTRY-COLSPEC-NAMEST-NODE)
  (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC" "colname"              
    (attribute-string "namest") (children (ancestor "tgroup"))))
(define (CALS-ENTRY-COLSPEC-NAMEST)
  (child-number (CALS-ENTRY-COLSPEC-NAMEST-NODE)))

; calculate the spane information from "namest" and "nameend" on
; the "entry"   
(define (CALS-ENTRY-COLSPEC-NAMEEND-NODE)
  (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC" "colname"              
    (attribute-string "nameend") (children (ancestor "tgroup"))))
(define (CALS-ENTRY-NAMEST-NAMEEND-NUMCOLS)
  (if (CALS-ENTRY-COLSPEC-NAMEEND-NODE)
     (+ 1
        (- (child-number (CALS-ENTRY-COLSPEC-NAMEEND-NODE))
           (child-number (CALS-ENTRY-COLSPEC-NAMEST-NODE))
     ))
     1))

; find the child number of the "colspec" that has a "colname" attribute
; value matching the "colname" attribute value of the "entry"
(define (CALS-ENTRY-COLSPEC-NAME-NODE)
  (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC" "colname"               
    (attribute-string "colname") (children (ancestor "tgroup"))))
(define (CALS-ENTRY-COLSPEC-NAME)
  (child-number (CALS-ENTRY-COLSPEC-NAME-NODE)))

; find the number of columns spanned by calculating the difference in child
; number of the "colspec"s that have a "colname" attribute whose value
; matches the "nameend" and "namest" attribute values of the "spanspec"
; whose "spannane" attribute value matches the "spanname" value of the "entry"
(define (CALS-ENTRY-SPANSPEC-NUMCOLS)
  (let ((spanspec-node 
          (GI-ATTVAL-NODE-IN-NODELIST "SPANSPEC" "spanname"
            (attribute-string "spanname") (children (ancestor "tgroup")))))
    (+ 1
       (- (child-number
            (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC"  "colname"
              (attribute-string "nameend" spanspec-node)
              (children (ancestor "tgroup"))))
          (child-number
            (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC"  "colname"
              (attribute-string "namest" spanspec-node)
              (children (ancestor "tgroup"))))))))

; find a "colspec" for a separator
(define (CALS-ENTRY-FIND-SEP-COLSPEC)
  (cond ((attribute-string "spanname") 
         (let ((spanspec-node 
               (GI-ATTVAL-NODE-IN-NODELIST "SPANSPEC" "spanname"
               (attribute-string "spanname") (children (ancestor "tgroup")))))
              (GI-ATTVAL-NODE-IN-NODELIST "COLSPEC"  "colname"
                (attribute-string "nameend" spanspec-node)
                (children (ancestor "tgroup")))))
        ((attribute-string "namest")   
         (CALS-ENTRY-COLSPEC-NAMEST-NODE))
        ((attribute-string "colname")  
         (CALS-ENTRY-COLSPEC-NAME-NODE))    
        (else #f)))

; find a "rowsep" specification by looking - in order - at
; "entry", "row", "colspec", "tgroup", "table"
(define (CALS-ENTRY-FIND-ROWSEP)
  (if (attribute-string "rowsep")
    (string->number (attribute-string "rowsep"))
    (if (attribute-string "rowsep" (ancestor "row"))
      (string->number (attribute-string "rowsep" (ancestor "row")))
      (let ((col-spec-node (CALS-ENTRY-FIND-SEP-COLSPEC)))
        (if (and col-spec-node (attribute-string "rowsep" col-spec-node))
          (string->number (attribute-string "rowsep" col-spec-node))
          (if (attribute-string "rowsep" (ancestor "tgroup"))
            (string->number (attribute-string "rowsep" (ancestor "tgroup")))
            (if (attribute-string "rowsep" (ancestor "table"))
              (string->number (attribute-string "rowsep" (ancestor "table")))
              %cals-rule-default%)))))))

; set up a value for the row separator - no distinction for values > 0
; if no rowsep found then use default separator
(define (CALS-ENTRY-ROWSEP)
  (let ((rowsep-value (CALS-ENTRY-FIND-ROWSEP)))
    (if (> rowsep-value 0)
	(if (last-sibling? (parent))
	    calc-table-head-body-border
	    calc-table-cell-after-row-border)
	#f)))

; find a "rowsep" specification by looking - in order - at
; current "colspec", "tgroup", "table"
(define (CALS-COLSPEC-FIND-ROWSEP)
  (if (attribute-string "rowsep")
    (string->number (attribute-string "rowsep"))
    (if (attribute-string "rowsep" (ancestor "tgroup"))
      (string->number (attribute-string "rowsep" (ancestor "tgroup")))
      (if (attribute-string "rowsep" (ancestor "table"))
        (string->number (attribute-string "rowsep" (ancestor "table")))
        %cals-rule-default%))))
; set up a value for the row separator - no distinction for values > 0
; if no rowsep found then use default separator
(define (CALS-COLSPEC-ROWSEP)
  (let ((rowsep-value (CALS-COLSPEC-FIND-ROWSEP)))
    (if (> rowsep-value 0)
	(if (last-sibling? (parent))
	    calc-table-head-body-border
	    calc-table-cell-after-row-border)
	#f)))

; find a "colsep" specification by looking - in order - at
; "entry", "colspec", "tgroup", "table"
(define (CALS-ENTRY-FIND-COLSEP)
  (if (attribute-string "colsep")
    (string->number (attribute-string "colsep"))
    (let ((col-spec-node (CALS-ENTRY-FIND-SEP-COLSPEC)))
      (if (and col-spec-node (attribute-string "colsep" col-spec-node))
        (string->number (attribute-string "colsep" col-spec-node))
        (if (attribute-string "colsep" (ancestor "tgroup"))
          (string->number (attribute-string "colsep" (ancestor "tgroup")))
          (if (attribute-string "colsep" (ancestor "table"))
            (string->number (attribute-string "colsep" (ancestor "table")))
            %cals-rule-default%))))))
; set up a value for the column separator - no distinction for values > 0
; if no colsep found then use default separator
(define (CALS-ENTRY-COLSEP)
  (let ((colsep-value (CALS-ENTRY-FIND-COLSEP)))
    (if (> colsep-value 0)
	calc-table-cell-after-column-border
       #f)))

; find a "colsep" specification by looking - in order - at
; current "colspec", "tgroup", "table"
(define (CALS-COLSPEC-FIND-COLSEP)
  (if (attribute-string "colsep")
    (string->number (attribute-string "colsep"))
    (if (attribute-string "colsep" (ancestor "tgroup"))
      (string->number (attribute-string "colsep" (ancestor "tgroup")))
      (if (attribute-string "colsep" (ancestor "table"))
        (string->number (attribute-string "colsep" (ancestor "table")))
        %cals-rule-default%))))
; set up a value for the row separator - no distinction for values > 0
; if no colsep found then use default separator
(define (CALS-COLSPEC-COLSEP)
  (let ((colsep-value (CALS-COLSPEC-FIND-COLSEP)))
    (if (> colsep-value 0)
	calc-table-cell-after-column-border
	#f)))

; find a "valign" specification by looking - in order - at
; "entry", "tbody", "thead", "tfoot"
(define (CALS-ENTRY-FIND-VALIGN)
  (if (attribute-string "valign")
    (string->number (attribute-string "valign"))
      (if (attribute-string "valign" (ancestor "tbody"))
        (string->number (attribute-string "valign" (ancestor "tbody")))
        (if (attribute-string "valign" (ancestor "thead"))
          (string->number (attribute-string "valign" (ancestor "thead")))
          (if (attribute-string "valign" (ancestor "tfoot"))
            (string->number (attribute-string "valign" (ancestor "tfoot")))
            %cals-valign-default%)))))
; set up a value for the row alignment
(define (CALS-ENTRY-VALIGN)
  (let ((valign-value (CALS-ENTRY-FIND-VALIGN)))
    (case valign-value
       (("TOP") 'start)
       (("MIDDLE") 'center)
       (("BOTTOM") 'end)
       (else 'start))))

(element TGROUP
  (let ((frame-attribute (inherited-attribute-string "frame")))
    (make table
      before-row-border:  (if (case frame-attribute
				(("ALL") #t)
				(("SIDES") #f)
				(("TOP") #t)
				(("BOTTOM") #f)
				(("TOPBOT") #t)
				(("NONE") #f)
				(else calc-table-before-row-border))
			      calc-table-before-row-border
			      #f)
      after-row-border:   (if (case frame-attribute
				(("ALL") #t)
				(("SIDES") #f)
				(("TOP") #f)
				(("BOTTOM") #t)
				(("TOPBOT") #t)
				(("NONE") #f)
				(else calc-table-after-row-border))
			      calc-table-after-row-border
			      #f)
      before-column-border: (if (case frame-attribute
				  (("ALL") #t)
				  (("SIDES") #t)
				  (("TOP") #f)
				  (("BOTTOM") #f)
				  (("TOPBOT") #f)
				  (("NONE") #f)
				  (else calc-table-before-column-border))
				calc-table-before-column-border
				#f)
      after-column-border:  (if (case frame-attribute
				  (("ALL") #t)
				  (("SIDES") #t)
				  (("TOP") #f)
				  (("BOTTOM") #f)
				  (("TOPBOT") #f)
				  (("NONE") #f)
				  (else calc-table-after-column-border))
				calc-table-after-column-border
				#f)
      (make table-part
	content-map: '((thead header)
		       (tbody #f)
		       (tfoot footer))
	(process-children)))))

(element COLSPEC 
  (make table-column
    cell-after-column-border: (CALS-COLSPEC-COLSEP)
    cell-after-row-border: (CALS-COLSPEC-ROWSEP)
    width: (CALS-COLSPEC-UNIT (attribute-string "colwidth"))))

(element THEAD
  (make sequence
    font-weight: 'bold
    label: 'thead))

(element TFOOT
  (make sequence
    label: 'tfoot))

(element TBODY
  (make sequence
    label: 'tbody))

(element ROW
  (let* ((tgroup (ancestor "TRGROUP"))
	 (table  (parent tgroup))
	 (rowsep (if (attribute-string "ROWSEP")
		     (attribute-string "ROWSEP")
		     (if (attribute-string "ROWSEP" tgroup)
			 (attribute-string "ROWSEP" tgroup)
			 (if (attribute-string "ROWSEP" table)
			     (attribute-string "ROWSEP" table)
			     (number->string %cals-rule-default%)))))
	 (after-row-border (> (string->number rowsep) 0)))
    (make table-row
      cell-after-row-border: (if after-row-border
				 (if (last-sibling?)
				     calc-table-head-body-border
				     calc-table-after-row-border)
				 #f)
      (process-children-trim))))

(element ENTRY
  (if (attribute-string "spanname")
      (make table-cell
	column-number: (CALS-ENTRY-SPANSPEC-START)
	n-columns-spanned: (CALS-ENTRY-SPANSPEC-NUMCOLS)
	n-rows-spanned: (+ 1 (let ((morerows (attribute-string "morerows")))
			       (if
				(string? morerows)
				(string->number morerows)
				0)))
	cell-row-alignment: (CALS-ENTRY-VALIGN)
	cell-after-column-border: (CALS-ENTRY-COLSEP)
	cell-after-row-border: (CALS-ENTRY-ROWSEP)
	cell-before-row-margin: %cals-cell-before-row-margin%
	cell-after-row-margin: %cals-cell-after-row-margin%
	cell-before-column-margin: %cals-cell-before-column-margin%
	cell-after-column-margin: %cals-cell-after-column-margin%
	start-indent: %cals-cell-content-start-indent%
	end-indent: %cals-cell-content-end-indent%
	(process-children-trim))
      (if (attribute-string "namest")
	  (make table-cell
            column-number: (CALS-ENTRY-COLSPEC-NAMEST)
            n-columns-spanned: (CALS-ENTRY-NAMEST-NAMEEND-NUMCOLS)
            n-rows-spanned: (+ 1 (let ((morerows (attribute-string "morerows")))
				   (if
				    (string? morerows)
				    (string->number morerows)
				    0)))
            cell-row-alignment: (CALS-ENTRY-VALIGN)
            cell-after-column-border: (CALS-ENTRY-COLSEP)
            cell-after-row-border: (CALS-ENTRY-ROWSEP)
            cell-before-row-margin: %cals-cell-before-row-margin%
            cell-after-row-margin: %cals-cell-after-row-margin%
            cell-before-column-margin: %cals-cell-before-column-margin%
            cell-after-column-margin: %cals-cell-after-column-margin%
            start-indent: %cals-cell-content-start-indent%
            end-indent: %cals-cell-content-end-indent%
            (process-children-trim))
	  (if (attribute-string "colname")
	      (make table-cell
		column-number: (CALS-ENTRY-COLSPEC-NAME)
		n-rows-spanned: (+ 1 (let ((morerows (attribute-string "morerows")))
				       (if
					(string? morerows)
					(string->number morerows)
					0)))
		cell-row-alignment: (CALS-ENTRY-VALIGN)
		cell-after-column-border: (CALS-ENTRY-COLSEP)
		cell-after-row-border: (CALS-ENTRY-ROWSEP)
		cell-before-row-margin: %cals-cell-before-row-margin%
		cell-after-row-margin: %cals-cell-after-row-margin%
		cell-before-column-margin: %cals-cell-before-column-margin%
		cell-after-column-margin: %cals-cell-after-column-margin%
		start-indent: %cals-cell-content-start-indent%
		end-indent: %cals-cell-content-end-indent%
		(process-children-trim))

	      (make table-cell
		n-rows-spanned: (+ 1 (let ((morerows (attribute-string "morerows")))
				       (if
					(string? morerows)
					(string->number morerows)
					0)))
		cell-row-alignment: (CALS-ENTRY-VALIGN)
		cell-after-column-border: (CALS-ENTRY-COLSEP)
		cell-after-row-border: (CALS-ENTRY-ROWSEP)
		cell-before-row-margin: %cals-cell-before-row-margin%
		cell-after-row-margin: %cals-cell-after-row-margin%
		cell-before-column-margin: %cals-cell-before-column-margin%
		cell-after-column-margin: %cals-cell-after-column-margin%
		start-indent: %cals-cell-content-start-indent%
		end-indent: %cals-cell-content-end-indent%
		(process-children-trim))))))

;; bosak 1996.11.23
(element (TABLE TITLE) (empty-sosofo)) ; don't show caption below table

(element SPANSPEC (empty-sosofo))

;; ===================== END OF TABLES ===============================

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Additional procedures

(define upperalpha
  '(#\A #\B #\C #\D #\E #\F #\G #\H #\I #\J #\K #\L #\M
    #\N #\O #\P #\Q #\R #\S #\T #\U #\V #\W #\X #\Y #\Z))

(define loweralpha
  '(#\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m
    #\n #\o #\p #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z))

(define (char-downcase ch)
  (case ch
    ((#\A) #\a) ((#\B) #\b) ((#\C) #\c) ((#\D) #\d) ((#\E) #\e)
    ((#\F) #\f) ((#\G) #\g) ((#\H) #\h) ((#\I) #\i) ((#\J) #\j)
    ((#\K) #\k) ((#\L) #\l) ((#\M) #\m) ((#\N) #\n) ((#\O) #\o)
    ((#\P) #\p) ((#\Q) #\q) ((#\R) #\r) ((#\S) #\s) ((#\T) #\t)
    ((#\U) #\u) ((#\V) #\v) ((#\W) #\w) ((#\X) #\x) ((#\Y) #\y)
    ((#\Z) #\z) (else ch)))

(define (LOCASE slist)
  (if (null? slist)
      '()
      (cons (char-downcase (car slist)) (LOCASE (cdr slist)))))

(define (STR2LIST s)
  (let ((len (string-length s)))
    (let loop ((i 0) (ln len))
      (if (= i len)
	  '()
	  (cons (string-ref s i) (loop (+ i 1) ln))))))

(define (STRING-DOWNCASE s)
  (apply string (LOCASE (STR2LIST s))))

(define (UNAME-START-INDEX u last)
  (let ((c (string-ref u last)))
    (if (or (member c upperalpha) (member c loweralpha))
	(if (= last 0)
	    0
	    (UNAME-START-INDEX u (- last 1)))
        (+ last 1))))

